from asq.initiators import query
from asq.selectors import k_

students = [dict(firstname='Joe', lastname='Blogs', scores=[56, 23, 21, 89]),
            dict(firstname='John', lastname='Doe', scores=[34, 12, 92, 93]),
            dict(firstname='Jane', lastname='Doe', scores=[33, 94, 91, 13]),
            dict(firstname='Ola', lastname='Nordmann', scores=[98, 23, 98, 87]),
            dict(firstname='Kari', lastname='Nordmann', scores=[86, 37, 88, 87]),
            dict(firstname='Mario', lastname='Rossi', scores=[37, 95, 45, 18])]

# Find all students where the first score is over 90
first_over_90 = query(students).where(lambda student: student['scores'][0] > 90)

# Select the last name of students where the first score is over 90
query(students).where(lambda student: student['scores'][0] > 90) \
               .select(lambda student: student['lastname'])

# Compute the maximum of all scores overall
query(students).select_many(lambda student: student['scores']).max()

# Order the students by last name and then by first name
query(students).order_by(lambda x: x['lastname']) \
               .then_by(lambda x: x['firstname'])


# It is very common to have selectors which retrieve a particular attribute
# from an object or index into the object using a particular key.  In these
# cases Python's lambda syntax can seem quite clunky.  To mitigate this, asq
# provides two factory functions called a_() and k_() which concisely generate
# selector functions for attributes and keys respectively.  The previous
# example could thus be rendered as::

query(students).order_by(k_('lastname')).then_by(k_('firstname'))



import heapq
import itertools
import functools
import multiprocessing

# Temporary warning
import sys
sys.stderr.write("Warning: The asq parallel query functionality should be "
                 "considered to be alpha quality.")

from .queryables import (Queryable, identity, default)


def star(func_and_args):
    func, args = func_and_args
    return func(*args)


class ParallelQueryable(Queryable):
    '''
    A parallel version of Queryable using the multiprocessing module.

    Warning: This parallel query functionality should be considered to be
        alpha quality.
    '''
    def __init__(self, iterable, pool=None, chunksize=1):
        super(ParallelQueryable, self).__init__(iterable)

        #TODO [asq 2.0] Support for shared pools

        self._own_pool = pool is None

        if self._own_pool:
            pool = multiprocessing.Pool()

        self._pool = pool
        self._chunksize = chunksize

    def _create(self, iterable):
        return ParallelQueryable(iterable, self._pool, self._chunksize)

    def _create_ordered(self, iterable, func=None):
        return OrderedParallelQueryable(iterable, func, self._pool,
                                        self._chunksize)

    def close(self):
        if self._own_pool:
            self._pool.close()
        super(ParallelQueryable, self).close()

    def select(self, selector):
        '''Transforms each element of a sequence into a new form.

        Each element is transformed through a selector function to produce a
        value for each value in the source sequence. The generated sequence is
        lazily evaluated.

        Args:
            selector: A unary function mapping a value in the source sequence
                to the corresponding value in the generated generated sequence.
                The argument of the selector function (which can have any name)
                is,

                Args:
                    element: The value of the element

                Returns:
                    The selected value derived from the element value

        Returns:
            A generated sequence whose elements are the result of invoking the
            selector function on each element of the source sequence.
        '''
        return self._create(self._pool.imap_unordered(selector, iter(self),
                                                      self._chunksize))

    def select_with_index(self, selector):
        '''Transforms each element of a sequence into a new form, incorporating
        the index of the element.

        Each element is transformed through a selector function which accepts
        the element value and its zero-based index in the source sequence. The
        generated sequence is lazily evaluated.

        Args:
            selector: A two argument function mapping the index of a value in
                the source sequence and the element value itself to the
                corresponding value in the generated sequence. The two
                arguments of the selector function (which can have any names)
                and its return value are,

                Args:
                    index: The zero-based index of the element
                    element: The value of the element

                Returns:
                    The selected value derived from the index and element

        Returns:
            A generated sequence whose elements are the result of invoking the
            selector function on each element of the source sequence
        '''

        return self._create(self._pool.imap_unordered(star,
                                    zip(itertools.repeat(selector),
                                      enumerate(iter(self))), self._chunksize))

    def select_many(self, projector, selector=identity):
        '''Projects each element of a sequence to an intermediate new sequence,
        flattens the resulting sequence into one sequence and optionally
        transforms the flattened sequence using a selector function.

        Args:
            projector: A unary function mapping each element of the source
                sequence into an intermediate sequence. If no projection
                function is provided, the intermediate sequence will consist of
                the single corresponding element from the source sequence. The
                projector function argument (which can have any name) and
                return values are,

                Args:
                    element: The value of the element

                Returns:
                    An iterable derived from the element value

            selector: An optional unary function mapping the elements in the
                flattened intermediate sequence to corresponding elements of
                the result sequence. If no selector function is provided, the
                identity function is used.  The selector function argument and
                return values are,

                Args:
                    element: The value of the intermediate element from the
                        concatenated sequences arising from the projector
                        function.

                Returns:
                    The selected value derived from the element value
        Returns:
            A generated sequence whose elements are the result of projecting
            each element of the source sequence using projector function and
            then mapping each element through an optional selector function.
        '''
        sequences = (self._create(item).select(projector) for item in iter(self))
        # TODO: [asq 2.0] Without the list() to force evaluation
        # multiprocessing deadlocks...
        chained_sequence = list(itertools.chain.from_iterable(sequences))
        return self._create(self._pool.imap_unordered(selector,
                                            chained_sequence, self._chunksize))

    # TODO: [asq 2.0] Replace lambda with a named module-scope function
    def select_many_with_index(self, projector=lambda i, x: [x], selector=identity):
        sequences = (self._create(item).select_with_index(projector) for item in iter(self))
        chained_sequence = itertools.chain.from_iterable(sequences)
        return self._create(self._pool.imap_unordered(selector, chained_sequence, self._chunksize))

    def select_many_with_correspondence(self, projector=lambda x: [x], selector=lambda x, y: y):
        corresponding_projector = lambda x: (x, projector(x))
        corresponding_selector = lambda x_y: selector(x_y[0], x_y[1])

        sequences = (Queryable(item).select(corresponding_projector) for item in iter(self))
        chained_sequence = itertools.chain.from_iterable(sequences)
        return self._create(self._pool.imap_unordered(corresponding_selector, chained_sequence, self._chunksize))

    def order_by(self, func=identity):
        return self._create_ordered(iter(self), func)

    # TODO: [asq 2.0] order_by_descending

    def where(self, predicate):
        partitions = realize_partitions(iter(self))
        filterer = functools.partial(filter, predicate)
        filtered_partitions = self._pool.imap_unordered(filterer, partitions, self._chunksize)
        return itertools.chain.from_iterable(filtered_partitions)

    def aggregate(self, func, seed=default):
        partitions = realize_partitions(iter(self))
        if len(partitions) == 1 and len(partitions[0]) == 1:
            if seed is default:
                return partitions[0][0]
            else:
                return func(seed, partitions[0][0])
        reducer = functools.partial(functools.reduce, func)
        reduced_partitions = self._pool.map_unordered(reducer, partitions, self._chunksize)
        return ParallelQueryable(reduced_partitions).aggregate(func, seed)

    def as_ordered(self):
        return self._create_ordered(iter(self))

    def as_sequential(self):
        import queryable
        return queryable.Queryable(iter(self))


class OrderedParallelQueryable(ParallelQueryable):
    '''A ParallelQueryable representing an ordered iterable.

    Warning: This parallel query functionality should be considered to be
        alpha quality.
    '''

    def __init__(self, iterable, func=None, pool=None, chunksize=1):
        super(OrderedParallelQueryable, self).__init__(iterable, pool, chunksize)
        self.funcs = [func] if func is not None else []

    def create(self, iterable):
        return OrderedParallelQueryable(iterable, self._pool, self._chunksize)

    def select(self, selector):
        '''Transforms each element of a sequence into a new form.

        Each element is transformed through a selector function to produce a
        value for each value in the source sequence. The generated sequence is
        lazily evaluated.

        Args:
            selector: A unary function mapping a value in the source sequence
                to the corresponding value in the generated generated sequence.
                The argument of the selector function (which can have any name)
                is,

                Args:
                    element: The value of the element

                Returns:
                    The selected value derived from the element value

        Returns:
            A generated sequence whose elements are the result of invoking the
            selector function on each element of the source sequence.
        '''
        return self.create(self._pool.imap(selector, iter(self), self._chunksize))

    def as_unordered(self):
        return self._create(iter(self))

    def then_by(self, func=identity):
        self.funcs.append(func)
        return self

    # TODO: [asq 2.0] order_by_descending, then_by_descending

    def __iter__(self):
        partitions = realize_partitions(self._iter())
        # TODO: [asq 2.0] Ensure that the sort is stable
        # TODO: [asq 2.0] Try using functools.partial and respond to
        # http://techguyinmidtown.com/2009/01/23/hack-for-functoolspartial-and-multiprocessing/
        # Actually, maybe functools.partial is pickleable in Python 3
        # http://www.mail-archive.com/python-bugs-list@python.org/msg47732.html
        sorted_partitions = self._pool.map(sorter, zip(itertools.repeat(self.funcs), partitions), self._chunksize)
        return heapq.merge(*sorted_partitions)


def sorter(funcs_iterable):
    funcs, iterable = funcs_iterable
    decorated = [(tuple(func(item) for func in funcs), item) for item in iterable]
    decorated.sort()
    undecorated = [item for funcs, item in decorated]
    return undecorated


def realize_partitions(iterable, floor=1, ceiling=32768):
    '''Partition the input sequence into a list of lists'''
    return [list(part) for part in geometric_partitions(iterable)]


def geometric_partitions(iterable, floor=1, ceiling=32768):
    '''
    Partition an iterable into chunks.  Returns an iterator over partitions.
    '''
    partition_size = floor
    run_length = multiprocessing.cpu_count()
    run_count = 0

    try:
        while True:
            #print("partition_size =", partition_size)
            # Split the iterable and replace the original iterator to avoid
            # advancing it
            partition, iterable = itertools.tee(iterable)

            # Yield the first partition, limited to the partition size
            yield Queryable(partition).take(partition_size)

            # Advance to the start of the next partition, this will raise
            # StopIteration if the iterator is exhausted
            for i in range(partition_size):
                next(iterable)

            # If we've reached the end of a run of this size, double the
            # partition size
            run_count += 1
            if run_count >= run_length:
                partition_size *= 2
                run_count = 0

            # Unless we have hit the ceiling
            if partition_size > ceiling:
                partition_size = ceiling

    except StopIteration:
        pass



# Asq's setup.py

from setuptools import setup
import io
import os
import re


def read(*names, **kwargs):
    with io.open(
        os.path.join(os.path.dirname(__file__), *names),
        encoding=kwargs.get("encoding", "utf8")
    ) as fp:
        return fp.read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")

here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the relevant file
with io.open(os.path.join(here, 'DESCRIPTION.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="asq",
    packages=["asq"],
    version=find_version("asq/version.py"),
    description="LINQ-for-objects style queries for Python iterables.",
    author="Robert Smallshire",
    author_email="robert@smallshire.org.uk",
    url="https://github.com/rob-smallshire/asq",
    keywords=["Python", "LINQ"],
    license="MIT License",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Environment :: Other Environment",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Utilities",
        ],
    long_description=long_description
)



import unittest
from asq.queryables import Queryable
from asq.test.test_queryable import TracingGenerator, infinite

__author__ = "Robert Smallshire"

class TestConcat(unittest.TestCase):

    def test_concat(self):
        a = [1, 2, 3]
        b = Queryable(a).concat([4, 5, 6]).to_list()
        c = [1, 2, 3, 4, 5, 6]
        self.assertEqual(b, c)

    def test_concat_infinite(self):
        b = Queryable(infinite()).concat(infinite()).take(3).to_list()
        c = [0, 1, 2]
        self.assertEqual(b, c)

    def test_concat_is_deferred(self):
        a = TracingGenerator()
        self.assertEqual(a.trace, [])
        b = TracingGenerator()
        self.assertEqual(b.trace, [])
        c = Queryable(a).concat(b).take().to_list()
        self.assertEqual(a.trace, [0])
        self.assertEqual(b.trace, [])

    def test_concat_non_iterable(self):
        a = [1, 2, 3, 4, 5]
        self.assertRaises(TypeError, lambda: Queryable(a).concat(7))

    def test_concat_closed(self):
        b = Queryable([1])
        b.close()
        self.assertRaises(ValueError, lambda: b.concat([2]))
  


import unittest
from asq.queryables import Queryable
from asq.test.test_queryable import infinite, TracingGenerator


class TestSelectWithIndex(unittest.TestCase):

    def test_select_with_index(self):
        a = [27, 74, 18, 48, 57, 97]
        b = Queryable(a).select_with_index().to_list()
        c = [(0, 27), (1, 74), (2, 18), (3, 48), (4, 57), (5, 97)]
        self.assertEqual(b, c)

    def test_select_with_index_finite(self):
        a = [27, 74, 18, 48, 57, 97, 76, 20, 91, 8, 80, 59, 20, 32, 58, 12, 74, 78, 4]
        b = Queryable(a).select_with_index(lambda x, y: x*y).to_list()
        c = [0, 74, 36, 144, 228, 485, 456, 140, 728, 72, 800, 649, 240, 416, 812, 180, 1184, 1326, 72]
        self.assertEqual(b, c)

    def test_select_with_index_infinite(self):
        a = infinite()
        b = Queryable(a).select_with_index(lambda x, y: x*y).take(4).to_list()
        c = [0, 1, 4, 9]
        self.assertEqual(b, c)

    def test_select_with_index_deferred(self):
        a = TracingGenerator()
        self.assertEqual(a.trace, [])
        b = Queryable(a).select_with_index()
        self.assertEqual(a.trace, [])
        b.take(3).to_list()
        self.assertEqual(a.trace, [0, 1, 2])

    def test_select_with_index_not_callable(self):
        a = [27, 74, 18, 48, 57, 97, 76, 20, 91, 8, 80, 59, 20, 32, 58, 12, 74, 78, 4]
        self.assertRaises(TypeError, lambda: Queryable(a).select_with_index("not callable"))

    def test_select_with_index_closed(self):
        a = [27, 74, 18, 48, 57, 97, 76, 20, 91, 8, 80, 59, 20, 32, 58, 12, 74, 78, 4]
        b = Queryable(a)
        b.close()
        self.assertRaises(ValueError, lambda: b.select_with_index(lambda x, y: x*y))





import sys
from timeit import Timer

from asq.initiators import query
from asq.selectors import a_

class Foo(object):
    def __init__(self):
        self.value = 42

items = []

def setup():
    for i in range(1000000):
        items.append(Foo())

def bench():
    results = query(items).select(a_("value")).to_list()

def main():
    setup()
    t = Timer("bench()", "from __main__ import bench")
    print(min(t.repeat(5, 1)))
    return 0

if __name__ == '__main__':
    sys.exit(main())


import unittest
from asq.queryables import Queryable
from asq.test.test_queryable import TracingGenerator, infinite

__author__ = "Robert Smallshire"

class TestIntersect(unittest.TestCase):

    def test_intersect(self):
        a = [1, 2, 3, 4, 5, 6, 7, 8]
        b = [2, 4, 9, 11]
        c = Queryable(a).intersect(b).to_list()
        d = [2, 4]
        self.assertEqual(c, d)

    def test_intersect_non_iterable(self):
        a = [1, 2, 3, 4, 5, 6, 7, 8]
        b = None
        self.assertRaises(TypeError, lambda: Queryable(a).intersect(b))

    def test_intersect_disjoint(self):
        a = [1, 2, 3, 4, 5]
        b = [6, 7, 8, 9, 10]
        c = Queryable(a).intersect(b).to_list()
        d = []
        self.assertEqual(c, d)

    def test_intersect_selector(self):
        a = [1, 2, 3, 4, -5, 6, 7, -8]
        b = [-2, -4, 5, 8]
        c = Queryable(a).intersect(b, abs).to_list()
        d = [2, 4, -5, -8]
        self.assertEqual(c, d)

    def test_intersect_selector_not_callable(self):
        a = [1, 2, 3, 4, -5, 6, 7, -8]
        b = [-2, -4, 5, 8]
        self.assertRaises(TypeError, lambda: Queryable(a).intersect(b, "not callable"))

    def test_intersect_infinite(self):
        b = [3, 7, 2, 9, 10]
        c = Queryable(infinite()).intersect(b).take(5).to_list()
        d = [2, 3, 7, 9, 10]
        self.assertEqual(c, d)

    def test_intersect_deferred(self):
        a = TracingGenerator()
        self.assertEqual(a.trace, [])
        b = [3, 7, 2, 9, 10]
        c = Queryable(a).intersect(b)
        self.assertEqual(a.trace, [])
        d = c.take(5).to_list()
        e = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        self.assertEqual(a.trace, e)

    def test_intersect_distinct(self):
        a = [1, 1, 2, 2, 4, 5, 3]
        b = [2, 5, 5]
        c = Queryable(a).intersect(b).to_list()
        d = [2, 5]
        self.assertEqual(c, d)

    def test_intersect_closed(self):
        a = [1, 1, 2, 2, 4, 5, 3]
        b = [2, 5, 5]
        b = Queryable(a)
        b.close()
        self.assertRaises(ValueError, lambda: b.intersect(b))


import unittest
from asq.queryables import Queryable

__author__ = "Robert Smallshire"

class TestToSet(unittest.TestCase):

    def test_to_set(self):
        a = [1, 2, 4, 8, 16, 32]
        b = Queryable(a).to_set()
        c = set([1, 2, 4, 8, 16, 32])
        self.assertEqual(b, c)

    def test_to_set_closed(self):
        a = [1, 2, 4, 8, 16, 32]
        b = Queryable(a)
        b.close()
        self.assertRaises(ValueError, lambda: b.to_set())

    def test_to_set_duplicates(self):
        a = [1, 2, 4, 8, 8, 16, 32]
        b = Queryable(a)
        self.assertRaises(ValueError, lambda: b.to_set())
